"""PyCXpress is a high-performance hybrid framework that seamlessly integrates Python and C++ to harness the flexibility of Python and the speed of C++ for efficient and expressive computation, particularly in the realm of deep learning and numerical computing."""

__all__ = [
    "TensorMeta",
    "ModelAnnotationCreator",
    "ModelAnnotationType",
    "ModelRuntimeType",
    "convert_to_spec_tuple",
    "pycxpress_debugger",
    "get_include",
    "version",
]

from importlib import metadata as importlib_metadata
from pathlib import Path


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


version: str = get_version()


from .core import (
    ModelAnnotationCreator,
    ModelAnnotationType,
    ModelRuntimeType,
    TensorMeta,
    convert_to_spec_tuple,
)
from .debugger import pycxpress_debugger


def get_include() -> str:
    return str(Path(__file__).parent.absolute() / "include")



# mypy: disable_error_code="type-arg,arg-type,union-attr,operator,assignment,misc"
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

from collections import namedtuple
from dataclasses import dataclass
from enum import Enum, auto
from itertools import chain

import numpy as np
from numpy.typing import DTypeLike

from .interface import (
    InputTensorProtocol,
    ModelProtocol,
    OutputTensorProtocol,
    TensorBufferProtocol,
)
from .utils import get_c_type, logger


@dataclass
class TensorMeta:
    dtype: DTypeLike  # the data type similar to np.int_
    shape: Union[
        int, Iterable[int], Callable[..., Union[int, Iterable[int]]]
    ]  # the maximal size of each dimension
    name: Optional[str] = None
    doc: Optional[str] = None

    def to_dict(self, *args, **kwargs) -> Dict:
        assert self.name is not None

        max_size = self.shape
        if isinstance(max_size, Callable):
            max_size = max_size(*args, **kwargs)
        max_size = tuple(max_size) if isinstance(self.shape, Iterable) else (max_size,)

        dtype, itemsize = get_c_type(self.dtype)

        return {
            "name": self.name,
            "dtype": dtype,
            "shape": tuple(round(-i) if i < 0 else None for i in max_size),
            "buffer_size": np.prod([round(abs(i)) for i in max_size]) * itemsize,
            "doc": f"" if self.doc is None else self.doc,
        }

    def setdefault(self, name: str) -> str:
        if self.name is None:
            self.name = name
        return self.name


class ModelAnnotationType(Enum):
    Input = auto()
    Output = auto()
    Operator = auto()
    HyperParams = auto()


class ModelRuntimeType(Enum):
    GraphExecution = auto()
    EagerExecution = auto()
    OfflineExecution = auto()


@dataclass
class TensorWithShape:
    data: Optional[np.ndarray] = None
    shape: Optional[Tuple] = None


class ModelAnnotationCreator(type):
    def __new__(
        mcs,
        name,
        bases,
        attrs,
        fields: Dict[str, TensorMeta],
        type: ModelAnnotationType,
        mode: ModelRuntimeType,
    ):
        if type == ModelAnnotationType.Input:
            generate_property = mcs.generate_input_property
        elif type == ModelAnnotationType.Output:
            generate_property = mcs.generate_output_property
        else:
            raise NotImplementedError()

        for field_name, field_meta in fields.items():
            field_meta.setdefault(field_name)
            attrs[field_name] = generate_property(field_meta)

        get_buffer_shape, set_buffer_value, init_func = mcs.general_funcs(
            name, [field_meta.name for field_meta in fields.values()]
        )

        attrs["__init__"] = init_func
        attrs["set_buffer_value"] = set_buffer_value
        if type == ModelAnnotationType.Output:
            attrs["get_buffer_shape"] = get_buffer_shape
        attrs.setdefault("__slots__", []).append("__buffer_data__")

        return super().__new__(mcs, name, bases, attrs)

    @staticmethod
    def general_funcs(name: str, field_names: List[str]):
        def get_buffer_shape(self, name: str) -> Tuple[int]:
            shape: Tuple[int] = getattr(self.__buffer_data__, name).shape
            return shape

        def set_buffer_value(self, name: str, value: np.ndarray) -> None:
            buffer = getattr(self.__buffer_data__, name)
            buffer.data = value

        def init_func(self):
            _BufferData_ = namedtuple("_BufferData_", field_names)
            self.__buffer_data__ = _BufferData_(
                *tuple(TensorWithShape() for _ in field_names)
            )

        return get_buffer_shape, set_buffer_value, init_func

    @staticmethod
    def generate_input_property(field: TensorMeta):
        def get_func(self):
            return getattr(self.__buffer_data__, field.name).data

        def set_func(*_):
            raise AssertionError("Not supported for input tensor")

        def del_func(_):
            raise AssertionError("Not supported for input tensor")

        return property(fget=get_func, fset=set_func, fdel=del_func, doc=field.doc)

    @staticmethod
    def generate_output_property(field: TensorMeta):
        def get_func(self):
            logger.warning(f"Only read the data field {field.name} in debugging mode")
            buffer = getattr(self.__buffer_data__, field.name)
            return buffer.data[: np.prod(buffer.shape)].reshape(buffer.shape)

        def set_func(self, data):
            buffer = getattr(self.__buffer_data__, field.name)
            buffer.shape = data.shape
            buffer.data[: np.prod(data.shape)] = data.flatten()

        def del_func(_):
            raise AssertionError("Not supported for output tensor")

        return property(fget=get_func, fset=set_func, fdel=del_func, doc=field.doc)


def convert_to_spec_tuple(
    inputFields: Iterable[TensorMeta], outputFields: Iterable
) -> Iterable[TensorBufferProtocol]:
    return chain.from_iterable(
        [
            (
                (v["name"], v["dtype"], v["buffer_size"], False)
                for v in [v.to_dict() for v in inputFields]
            ),
            (
                (v["name"], v["dtype"], v["buffer_size"], True)
                for v in [v.to_dict() for v in outputFields]
            ),
        ]
    )


def main():
    pass


if __name__ == "__main__":
    main()



# mypy: disable_error_code="type-arg"
from typing import List, Protocol, Tuple

from abc import abstractmethod

import numpy as np


class InputTensorProtocol(Protocol):
    def __init__(self, *args, **kwargs): ...

    @abstractmethod
    def set_buffer_value(self, name: str, value: np.ndarray) -> None: ...


class OutputTensorProtocol(Protocol):
    def __init__(self, *args, **kwargs): ...

    @abstractmethod
    def get_buffer_shape(self, name: str) -> Tuple[int]: ...

    @abstractmethod
    def set_buffer_value(self, name: str, value: np.ndarray) -> None: ...


TensorBufferProtocol = Tuple[
    str,  # name
    str,  # dtype
    int,  # buffer size
    bool,  # is_output
]


class ModelProtocol(Protocol):
    def __init__(self, *args, **kwargs): ...

    @abstractmethod
    def initialize(
        self,
    ) -> Tuple[
        InputTensorProtocol, OutputTensorProtocol, Tuple[TensorBufferProtocol]
    ]: ...

    @abstractmethod
    def run(self) -> None: ...

    @abstractmethod
    def reset(self) -> None: ...



# pylint: disable=missing-function-docstring

import argparse
import sys
import sysconfig

from pybind11 import get_include as pybind11_include

from PyCXpress import get_include, version


def print_includes() -> None:
    dirs = {
        sysconfig.get_path("include"),
        sysconfig.get_path("platinclude"),
        pybind11_include(),
        get_include(),
    }

    print(" ".join(f"-I {d}" for d in dirs))


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version",
        action="version",
        version=version,
        help="Print the version and exit.",
    )
    parser.add_argument(
        "--includes",
        action="store_true",
        help="Include flags for both pybind11 and Python headers.",
    )
    args = parser.parse_args()
    if not sys.argv[1:]:
        parser.print_help()
    if args.includes:
        print_includes()


if __name__ == "__main__":
    main()



from typing import Optional

import os

from .utils import logger

_debugger_status_ = [False]


def pycxpress_debugger(
    host: Optional[str] = None,
    port: Optional[int] = None,
    debugger: Optional[str] = None,
):
    if _debugger_status_[0] == True:
        return

    if debugger is None:
        return

    if host is None:
        host = os.environ.get("PYCXPRESS_DEBUGGER_HOST", "localhost")

    if port is None:
        port = int(os.environ.get("PYCXPRESS_DEBUGGER_PORT", "5678"))

    if debugger.lower() == "pycharm":
        try:
            import pydevd_pycharm

            pydevd_pycharm.settrace(
                host, port=port, stdoutToServer=True, stderrToServer=True, suspend=True
            )
            _debugger_status_[0] = True

        except ConnectionRefusedError:
            logger.warning(
                "Can not connect to Python debug server (maybe not started?)"
            )
            logger.warning(
                "Use PYCXPRESS_DEBUGGER_TYPE=debugpy instead as Pycharm professional edition is needed for Python debug server feature."
            )
    elif debugger.lower() == "debugpy":
        import debugpy

        _debugger_status_[0] = True

        debugpy.listen((host, port))
        logger.info(f"debugpy listen on {host}:{port}, please use VSCode to attach")
        debugpy.wait_for_client()
    else:
        logger.warning(
            f"Only PYCXPRESS_DEBUGGER_TYPE=debugpy|pycharm supported but {debugger} provided"
        )



from typing import Tuple

import logging

import numpy as np
from numpy.typing import DTypeLike

logger = logging.getLogger("PyCXpress")


def get_c_type(t: DTypeLike) -> Tuple[str, int]:
    dtype = np.dtype(t)
    relation = {
        np.dtype("bool"): "bool",
        np.dtype("int8"): "int8_t",
        np.dtype("int16"): "int16_t",
        np.dtype("int32"): "int32_t",
        np.dtype("int64"): "int64_t",
        np.dtype("uint8"): "uint8_t",
        np.dtype("uint16"): "uint16_t",
        np.dtype("uint32"): "uint32_t",
        np.dtype("uint64"): "uint64_t",
        np.dtype("float32"): "float",
        np.dtype("float64"): "double",
    }
    return relation.get(dtype, "char"), dtype.itemsize or 1



# mypy: disable_error_code="arg-type,type-arg,attr-defined"
import os

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
import logging

logging.basicConfig(level=logging.DEBUG)

import os
from contextlib import nullcontext

import numpy as np

from PyCXpress import (
    ModelAnnotationCreator,
    ModelAnnotationType,
    ModelRuntimeType,
    TensorMeta,
    convert_to_spec_tuple,
    pycxpress_debugger,
)

if "PYCXPRESS_DEBUGGER_TYPE" in os.environ:
    pycxpress_debugger(debugger=os.environ["PYCXPRESS_DEBUGGER_TYPE"])


def show(a: np.ndarray):
    logging.info(f"array data type: {a.dtype}")
    logging.info(f"array data shape: {a.shape}")
    logging.info(f"array data: ")
    logging.info(a)


InputFields = dict(
    data_to_be_reshaped=TensorMeta(
        dtype=np.float_,
        shape=(100,),
    ),
    new_2d_shape=TensorMeta(
        dtype=np.uint8,
        shape=-2,
    ),
)


class InputDataSet(
    metaclass=ModelAnnotationCreator,
    fields=InputFields,
    type=ModelAnnotationType.Input,
    mode=ModelRuntimeType.EagerExecution,
):
    pass


OutputFields = dict(
    output_a=TensorMeta(
        dtype=np.float_,
        shape=(10, 10),
    ),
)


class OutputDataSet(
    metaclass=ModelAnnotationCreator,
    fields=OutputFields,
    type=ModelAnnotationType.Output,
    mode=ModelRuntimeType.EagerExecution,
):
    pass


class Model:
    def __init__(self):
        self.input = None
        self.output = None

    def initialize(self):
        self.input, self.output = InputDataSet(), OutputDataSet()

        return (
            self.input,
            self.output,
            tuple(convert_to_spec_tuple(InputFields.values(), OutputFields.values())),
        )

    def run(self):
        self.model(self.input, self.output)

    @staticmethod
    def model(input: InputDataSet, output: OutputDataSet):
        with nullcontext():
            # print(input.data_to_be_reshaped)
            # print(input.new_2d_shape)
            output.output_a = input.data_to_be_reshaped.reshape(input.new_2d_shape).T
            # print(output.output_a)


def main():

    model = Model()
    input_data, output_data, spec = model.initialize()
    print(spec)

    input_data.set_buffer_value("data_to_be_reshaped", np.arange(12, dtype=np.float_))
    print(input_data.data_to_be_reshaped)
    input_data.set_buffer_value("new_2d_shape", np.array([3, 4]).astype(np.uint8))
    print(input_data.new_2d_shape)
    output_data.set_buffer_value("output_a", np.arange(12) * 0)

    model.run()
    print(output_data.output_a)
    print(output_data.get_buffer_shape("output_a"))


if __name__ == "__main__":
    main()
